@@ -1,38 +1,41 @@
-from operator import itemgetter
-from collections import defaultdict
 from random import Random
+from .context import requirements, consume
 
 
-class OperationOrLanguage(object):
-    precondition = None
-
-    def applicable(self, context):
-        for varstack, req in self.requirements.items():
-            if not context.varstack(varstack).has(req):
-                return False
-        if self.precondition:
-            return self.precondition(*context.read(self.argspec))
-        else:
-            return True
-
-
-class Operation(OperationOrLanguage):
+class Operation(object):
     def __init__(
-            self, varstacks, name=None, pattern=None, patterns=None,
-            precondition=None):
+            self, argspec, target=None, targets=None, name=None, pattern=None,
+            patterns=None):
         if pattern and patterns:
             raise ValueError(
                 "Cannot specify both a single and multiple patterns"
             )
         if pattern:
             patterns = (pattern,)
-        self.varstacks = tuple(map(itemgetter(0), varstacks))
+        if target and targets:
+            raise ValueError(
+                "Cannot specify both a single and multiple targets"
+            )
+        if target is not None:
+            self.single_target = True
+            targets = (target,)
+        else:
+            self.single_target = False
+        if targets is None:
+            targets = ()
+        if len(targets) == 1:
+            self.target = targets[0]
+        self.argspec = argspec
+        self.targets = targets
         self.name = name or self.__class__.__name__.lower()
-        self.requirements = defaultdict(lambda: 0)
-        for s, c in varstacks:
-            self.requirements[s] += c
+        self.requirements = requirements(argspec)
         self.patterns = patterns
-        self.precondition = precondition
+
+    def applicable(self, heights):
+        for varstack, req in self.requirements.items():
+            if heights.get(varstack, 0) < req:
+                return False
+        return True
 
     def __repr__(self):
         return "Operation(%s)" % self.display()
@@ -40,7 +43,9 @@ class Operation(OperationOrLanguage):
     def compile(self, arguments, results):
         if self.patterns is None:
             invocation = "{0}({1})".format(self.name, ', '.join(arguments))
-            return ["%s = %s" % (', '.join(results), invocation)]
+            if results:
+                invocation = "%s = %s" % (', '.join(results), invocation)
+            return [invocation]
         else:
             patterns = self.patterns
             if results:
@@ -52,7 +57,7 @@ class Operation(OperationOrLanguage):
             ]
 
     def args(self):
-        return self.varstacks
+        return ()
 
     def display(self):
         return "%s(%s)" % (
@@ -60,34 +65,23 @@ class Operation(OperationOrLanguage):
         )
 
     def invoke(self, context):
-        raise NotImplementedError()
+        raise NotImplementedError("%s.invoke(context)" % (self.__class__,))
+
+    def simulate(self, context):
+        context.read(self.argspec)
+        for t in self.targets:
+            context.varstack(t).push(None)
 
 
 class ReadAndWrite(Operation):
     def __init__(
-        self, function, argspec, target=None, targets=None, name=None,
-        precondition=None, pattern=None, patterns=None,
+        self, function, **kwargs
     ):
-        if target and targets:
-            raise ValueError(
-                "Cannot specify both a single and multiple targets"
-            )
-        if target is not None:
-            targets = (target,)
-            self.single_target = True
-        else:
-            self.single_target = False
+        if 'name' not in kwargs:
+            kwargs['name'] = function.__name__
 
-        super(ReadAndWrite, self).__init__(
-            _counts(argspec),
-            name or function.__name__,
-            patterns=patterns,
-            pattern=pattern,
-            precondition=precondition,
-        )
+        super(ReadAndWrite, self).__init__(**kwargs)
         self.function = function
-        self.argspec = argspec
-        self.targets = targets
 
     def invoke(self, context):
         args = context.read(self.argspec)
@@ -101,27 +95,25 @@ class ReadAndWrite(Operation):
                     context.varstack(target).push(v)
 
 
-class Mutate(Operation):
+class Mutate(ReadAndWrite):
     def __init__(
-        self, function, argspec, name=None, precondition=None,
+        self, function, argspec, name=None,
         pattern=None
     ):
-        super(ReadAndWrite, self).__init__(
-            _counts(argspec),
-            name or function.__name__,
+        super(Mutate, self).__init__(
+            argspec=argspec,
+            name=name or function.__name__,
             pattern=pattern,
-            precondition=precondition,
         )
 
 
 class BinaryOperator(ReadAndWrite):
-    def __init__(self, operation, varstack, name, precondition=None):
+    def __init__(self, operation, varstack, name):
         super(BinaryOperator, self).__init__(
             function=operation,
             argspec=(varstack, varstack),
             target=varstack,
             name=name,
-            precondition=precondition,
         )
 
     def compile(self, arguments, results):
@@ -164,7 +156,7 @@ class Check(Operation):
             pattern = "assert %s(%s)" % (name, arg_string)
 
         super(Check, self).__init__(
-            _counts(argspec), name=name, pattern=pattern, patterns=patterns
+            argspec=argspec, name=name, pattern=pattern, patterns=patterns
         )
         self.argspec = argspec
         self.test = test
@@ -174,59 +166,48 @@ class Check(Operation):
         assert self.test(*args)
 
 
-class SingleStackOperation(Operation):
-    min_height = 0
-
-    def __init__(self, varstack, name=None):
-        super(SingleStackOperation, self).__init__(
-            varstacks=((varstack, self.min_height),), name=name
-        )
-
-    @property
-    def varstack(self):
-        return self.varstacks[0]
-
-
-class Push(SingleStackOperation):
+class Push(Operation):
     def __init__(self, varstack, gen_value, value_formatter=None):
-        super(Push, self).__init__(varstack, name="push")
+        super(Push, self).__init__(argspec=(), target=varstack, name="push")
         self.gen_value = gen_value
         self.value_formatter = value_formatter or repr
 
     def compile(self, arguments, results):
         assert not arguments
-        assert len(results) == 1
+        assert len(results) <= 1
         v = self.gen_value()
         return [
             "%s = %s" % (results[0], self.value_formatter(v))
         ]
 
-    def invoke(self, context):
-        context.varstack(self.varstack).push(self.gen_value())
-
     def args(self):
-        return super(Push, self).args() + (self.gen_value(),)
+        return (self.gen_value(),)
+
+    def invoke(self, context):
+        context.varstack(self.target).push(self.gen_value())
 
 
 class InapplicableLanguage(Exception):
     pass
 
 
-class Language(OperationOrLanguage):
-    def __init__(self):
-        self.requirements = {}
-
-    def generate(self, context):
+class Language(object):
+    def generate(self, heights, random):
         """
-        Given a runcontext Context produce an Operation or raise
-        InapplicableLanguage if this Language cannot produce Operations in the
-        current state of context.
+        :param heights: A dict of string names to integer counts of the number
+            of values available each varstack
+        :param a random number generator
 
-        This should always throw InapplicableLanguage if the requirements are
-        not satisfied. It may throw it in other circumstances.
+        returns an Operation or raises InapplicableLanguage. Should not depend
+        on any factors other than its parameters. Whether it raises an
+        InapplicableLanguage should not depend on the state of the random
+        number generator.
         """
         raise NotImplementedError()
 
+    def generate_from(self, context):
+        return self.generate(context.heights(), context.random)
+
 
 class PushRandom(Language):
     def __init__(self, produce, target, name=None, value_formatter=None):
@@ -235,8 +216,8 @@ class PushRandom(Language):
         self.target = target
         self.value_formatter = value_formatter
 
-    def generate(self, context):
-        state = context.random.getstate()
+    def generate(self, heights, random):
+        state = random.getstate()
 
         def gen_result():
             r = Random(0)
@@ -264,48 +245,59 @@ class ChooseFrom(Language):
             adjusted.append(c)
         children = tuple(adjusted)
         self.children = children
-        self.requirements = defaultdict(lambda: 0)
-        for c in children:
-            for k, v in c.requirements.items():
-                self.requirements[k] = min(v, self.requirements[k])
 
-    def generate(self, context):
+    def generate(self, heights, random):
         children = list(self.children)
-        context.random.shuffle(children)
+        random.shuffle(children)
         for child in children:
-            if isinstance(child, Language):
-                try:
-                    return child.generate(context)
-                except InapplicableLanguage:
-                    continue
-            else:
-                if child.applicable(context):
-                    return child
+            try:
+                if isinstance(child, Operation):
+                    operation = child
+                else:
+                    operation = child.generate(heights, random)
+                if operation.applicable(heights):
+                    return operation
+            except InapplicableLanguage:
+                continue
         raise InapplicableLanguage
 
 
-class Dup(SingleStackOperation):
-    min_height = 1
+class Drop(Operation):
+    def __init__(self, varstack):
+        super(Drop, self).__init__(
+            argspec=(consume(varstack),)
+        )
+
+    def invoke(self, context):
+        context.read(self.argspec)
+
+    def compile(self, arguments, results):
+        return []
+
+
+class DataShufflingOperation(Operation):
+    def __init__(self, varstack):
+        super(DataShufflingOperation, self).__init__(
+            argspec=(varstack,) * self.height
+        )
+        self.varstack = varstack
+
+    def simulate(self, context):
+        return self.invoke(context)
+
+    def compile(self, arguments, results):
+        return []
+
+
+class Dup(DataShufflingOperation):
+    height = 1
 
     def invoke(self, context):
         context.varstack(self.varstack).dup()
 
-    def compile(self, arguments, results):
-        return []
 
-
-class Drop(SingleStackOperation):
-    min_height = 1
-
-    def invoke(self, context):
-        context.varstack(self.varstack).pop()
-
-    def compile(self, arguments, results):
-        return []
-
-
-class Swap(SingleStackOperation):
-    min_height = 2
+class Swap(DataShufflingOperation):
+    height = 2
 
     def invoke(self, context):
         vs = context.varstack(self.varstack)
@@ -314,12 +306,9 @@ class Swap(SingleStackOperation):
         vs.push(x)
         vs.push(y)
 
-    def compile(self, arguments, results):
-        return []
 
-
-class Rot(SingleStackOperation):
-    min_height = 3
+class Rot(DataShufflingOperation):
+    height = 3
 
     def invoke(self, context):
         vs = context.varstack(self.varstack)
@@ -329,13 +318,3 @@ class Rot(SingleStackOperation):
         vs.push(y)
         vs.push(z)
         vs.push(x)
-
-    def compile(self, arguments, results):
-        return []
-
-
-def _counts(values):
-    c = defaultdict(lambda: 0)
-    for v in values:
-        c[v] += 1
-    return c.items()
